#include "Util.h"

#include "clang/AST/ASTContext.h"
#include "clang/AST/Expr.h"
#include "clang/AST/Mangle.h"
#include "clang/Tooling/Refactoring.h"

#include "clang/Lex/Lexer.h"

#include "spdlog/spdlog.h"

using namespace clang;

std::string get_decl_name(const NamedDecl *d, ASTContext &c) {
  // Types do not have a mangled name
  if (const auto *type_decl = dyn_cast<clang::TypeDecl>(d)) {
    return type_decl->getNameAsString();
  }
  // We don't refer to variables with mangled names
  if (const auto *var_decl = dyn_cast<clang::VarDecl>(d)) {
    return var_decl->getNameAsString();
  }
  ASTNameGenerator name_gen{c};
  return name_gen.getName(d);
}

std::string get_decl_name(const NamedDecl *d) {
  assert(d);
  return get_decl_name(d, d->getASTContext());
}

std::string stmt2str(const Stmt *st, SourceManager &sm) {
  std::string str = "[";
  std::string filename = sm.getFilename(st->getBeginLoc()).str();
  str.append(filename.substr(filename.rfind('/') + 1));
  str.append(":");
  str.append(std::to_string(sm.getSpellingLineNumber(st->getBeginLoc())));
  str.append("] ");
  str.append(Lexer::getSourceText(CharSourceRange::getTokenRange(st->getSourceRange()), sm, LangOptions(), nullptr));
  return str;
}

void triml(std::string &str) {
  str.erase(str.begin(), std::find_if(str.begin(), str.end(), [](auto c) { return !std::isspace(c); }));
}

void trimr(std::string &str) {
  str.erase(std::find_if(str.rbegin(), str.rend(), [](auto c) { return !std::isspace(c); }).base(), str.end());
}

std::string trim(std::string str) {
  triml(str);
  trimr(str);
  return str;
}

clang::SourceRange get_include_in_main_file(const SourceRange &loc, const SourceManager &sm) {
  if (loc.isInvalid()) {
    return loc;
  }
  auto include_loc = sm.getIncludeLoc(sm.getFileID(loc.getBegin()));
  if (include_loc.isInvalid()) {
    return include_loc;
  }
  const auto main_file_id = sm.getMainFileID();
  clang::SourceRange range(include_loc);
  if (sm.getFileID(include_loc) != main_file_id) {
    return get_include_in_main_file(range, sm);
  }
  return range;
}

clang::QualType get_underlying_type(QualType t) {
  if (const auto array = dyn_cast<clang::ArrayType>(t)) {
    return array->getElementType();
  } else if (const auto ptr = dyn_cast<clang::PointerType>(t)) {
    return ptr->getPointeeType();
  } else if (const auto incomplet_array = dyn_cast<clang::IncompleteArrayType>(t)) {
    return incomplet_array->getElementType();
  }
  llvm_unreachable("Unexpected type");
}

bool is_potential_mod(const clang::QualType &t) {
  return t->isPointerType() && !t->getPointeeType().isConstQualified();
}
